import math

from cobweb.defaults import MAX_FILTER_SIZE


class BloomFilter:

    def __init__(self, max_elements=MAX_FILTER_SIZE):
        self.maximum = max_elements
        self.size = 0
        self.items = 0

    def __contains__(self, item: str):
        _hash = hash(item)
        return _hash & self.items == _hash

    def __len__(self):
        return self.size

    def count_error_probability(self):
        return 2 ** - math.log(self.maximum / self.size) if self.size else 0

    def add(self, item: str):
        _hash = hash(item)
        self.size += 1
        self.items |= _hash

    def clear(self):
        self.items = 0
        self.size = 0



from PIL import Image

from cobweb.adapters.file import FileAdapter
from cobweb.bbs import BlockBypassSystem


class ImageAdapter(FileAdapter):

    bbs = BlockBypassSystem()

    @bbs.monitor_responses
    def process_response(self, response):
        response = super().process_response(response)
        response.processed_by = self.__class__

        response.data = Image.open(response.data)
        return response



import logging
from multiprocessing import cpu_count, Queue
from types import FunctionType

import urllib3

from cobweb.adapters.http import HttpAdapter
from cobweb.helpers import Singleton
from cobweb.link_extractor import LinkExtractor
from cobweb.settings import Settings
from cobweb.spiders import Spider


class Cobweb(metaclass=Singleton):

    def __init__(
            self,
            start_url=None,
            settings=Settings(),
            adapter=HttpAdapter(),
            link_extractor=LinkExtractor(),
            log_to_file: bool = False,
            logger_filename: str = None,
            file_logger_level=logging.INFO,
            console_logger_level=logging.INFO,
    ):
        self.start_url = start_url
        self._settings = settings
        self._spiders = None
        self._queue = Queue()
        self._adapter = adapter
        self._link_extractor = link_extractor
        self._default_process_function = None

        self.log_format = '[%(processName)s: %(asctime)s] %(levelname)-8s %(message)s'
        self.log_to_file = log_to_file
        self.logger_filename = logger_filename
        self.file_logger_level = file_logger_level
        self.console_logger_level = console_logger_level

    @property
    def logger(self):
        return logging.getLogger("cobweb")

    @property
    def settings(self):
        return self._settings

    @property
    def queue(self):
        return self._queue

    @property
    def adapter(self):
        return self._adapter

    @property
    def link_extractor(self):
        return self._link_extractor

    @link_extractor.setter
    def domain(self, value):
        self._link_extractor = value

    @property
    def default_process_function(self):
        return self._default_process_function

    @default_process_function.setter
    def default_process_function(self, value):
        if not isinstance(value, FunctionType) and value is not None:
            raise ValueError("Trying to set not a function to callback")
        else:
            self._default_process_function = value

    @property
    def domain_only(self):
        return self._link_extractor.domain_only

    @domain_only.setter
    def domain_only(self, value):
        self._link_extractor.domain_only = value

    def __getitem__(self, item: int):
        return self._spiders[item]

    def follow_rules(self, rules):
        self._link_extractor.follow_patterns = rules

    def exclude_rules(self, rules):
        self._link_extractor.exclude_patterns = rules

    def setup_logging(self) -> None:
        """
        method for configuring logging in all application
        :return: None
        """

        logger = self.logger
        logger.setLevel(self.console_logger_level)
        formatter = logging.Formatter(self.log_format)

        if self.log_to_file:
            if not self.logger_filename:
                raise ValueError("'log_to_file' was on but no 'logger_filename' specified")
            file_handler = logging.FileHandler('spam.log')
            file_handler.setLevel(logging.DEBUG)
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)

        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.ERROR)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

        logging.basicConfig(format=self.log_format)

    def start_from(self, url: str):
        self.start_url = url
        self.queue.put(self.start_url)

    def get_spider(self):
        if self.default_process_function is None:
            raise ValueError("No process function set")
        spider = Spider(
            queue=self.queue,
            settings=self.settings,
            adapter=self.adapter,
            link_extractor=self.link_extractor
        )
        spider.process_response = self.default_process_function
        return spider

    def populate_spiders(self, num=cpu_count()):
        self._spiders = [
            self.get_spider()
            for _ in range(num)
        ]

    def add_spider(self, spider: Spider):
        spider.queue = self._queue
        self._spiders.append(spider)

    def start(self):
        if not self.start_url:
            raise ValueError("No start url specified")

        self.link_extractor.urls_queue = self.queue

        self.setup_logging()
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        self.logger.info("starting spiders")

        self.logger.debug("on start queue size: %s", self.queue.qsize())

        for spider in self._spiders:
            spider.start()

        self.link_extractor.start()

    def stop(self):
        self.logger.info("stopping spiders")
        for spider in self._spiders:
            spider.stop()



import fnmatch
import logging
import multiprocessing
import re
from collections import Iterable
from queue import Empty
from typing import Union
from urllib.parse import urlparse, ParseResult

from bs4 import BeautifulSoup
from pygtrie import PrefixSet

from cobweb.link_extractor.bloom_filter import BloomFilter


class LinkExtractor(multiprocessing.Process):

    REGEX_UNIX = "unix"
    REGEX_REGULAR = "regular"

    PROBABILISTIC = "probabilistic"
    TRIE = "trie"
    HASH = "hash"

    def __init__(
            self,
            urls_queue=None,
            follow: Union[Iterable, str] = ("*",),
            exclude: Union[Iterable, str] = tuple(),
            regex_type=REGEX_UNIX,
            mode=HASH,
            domain_only=True
    ):
        multiprocessing.Process.__init__(self)

        self._current_page_url = None
        self._follow_patterns = None
        self._exclude_patterns = None

        self.domain_only = domain_only
        self.allowed_schemes = {
            "http",
            "https",
        }

        self.regex_type = regex_type
        self.follow_patterns = follow
        self.exclude_patterns = exclude

        self.urls_queue = urls_queue
        self.responses_queue = multiprocessing.Queue()

        self.mode = mode
        # Any object with both '__contains__' and 'add' methods defined can be used here
        if mode == LinkExtractor.PROBABILISTIC:
            self.seen = BloomFilter()
        elif mode == LinkExtractor.HASH:
            self.seen = set()
        elif mode == LinkExtractor.TRIE:
            self.seen = PrefixSet()

    @property
    def logger(self):
        return logging.getLogger("cobweb.link_extractor")

    @property
    def current_page_url(self):
        return self._current_page_url

    @current_page_url.setter
    def current_page_url(self, value: str):
        self._current_page_url = urlparse(value)

    @property
    def follow_patterns(self):
        return self._follow_patterns

    @follow_patterns.setter
    def follow_patterns(self, value):
        follow_patterns = value if not isinstance(value, str) else [value]
        self._follow_patterns = [
            self.get_regex(pattern)
            for pattern in follow_patterns
        ]

    @property
    def exclude_patterns(self):
        return self._exclude_patterns

    @exclude_patterns.setter
    def exclude_patterns(self, value):
        exclude_patterns = value if not isinstance(value, str) else [value]
        self._exclude_patterns = [
            self.get_regex(pattern)
            for pattern in exclude_patterns
        ]

    def get_regex(self, string: str):
        if self.regex_type == self.REGEX_UNIX:
            return re.compile(fnmatch.translate(string))
        else:
            return re.compile(string)

    def is_valid_domain(self, href: str) -> bool:
        """
        checks whether the given string satisfy the 'domain_only' setting
        :param href: str
        :return: bool
        """
        parsed = urlparse(href)
        if parsed.scheme and parsed.scheme not in self.allowed_schemes:
            return False
        if self.domain_only:
            return not parsed.netloc or parsed.netloc == self.current_page_url.netloc
        else:
            return parsed.scheme and parsed.netloc

    def match_follow_patterns(self, link):
        return any([
            pattern.match(link)
            for pattern in self.follow_patterns
        ])

    def match_exclude_patterns(self, link):
        return any([
            pattern.match(link)
            for pattern in self.exclude_patterns
        ])

    def is_satisfying_link(self, link: str) -> bool:
        """
        checks whether the given link satisfy 'follow' and 'exclude' preferences
        :param link: str
        :return: bool
        """
        return self.match_follow_patterns(link) and not self.match_exclude_patterns(link)

    def get_full_url(self, href: ParseResult):
        if not href.scheme:
            href = href._replace(scheme=self.current_page_url.scheme)
        if not href.netloc:
            href = href._replace(netloc=self.current_page_url.netloc)
        if href.path.startswith("./"):
            href = href._replace(path=href.path[2:])
        return href.geturl()

    def filter_valid_links(self, all_links: set):
        valid_links = set()
        for link in all_links:
            if link.has_attr("href") and self.is_valid_domain(link["href"]):
                url = self.get_full_url(urlparse(link["href"]))
                if self.is_satisfying_link(url):
                    valid_links.add(url)
                    self.seen.add(url)
        self.logger.debug("valid links count: %s", len(valid_links))
        if self.mode == self.PROBABILISTIC:
            self.logger.info("current error probability %s", self.seen.count_error_probability())
        return valid_links

    def extract(self, response):
        self.current_page_url = response.url
        data = BeautifulSoup(response.content, "html.parser")
        all_links = data.find_all("a")
        self.logger.debug("found %s new links", len(all_links))
        return self.filter_valid_links(all_links)

    def run(self):
        if not self.urls_queue:
            raise ValueError("No queue to take url from")
        while True:
            try:
                response = self.responses_queue.get()
            except Empty:
                pass
            else:
                for link in self.extract(response):
                    self.urls_queue.put(link)



import os

import requests
import time

USER_AGENT_PATH = os.path.join(
    *os.path.split(__file__)[:-1],
    "bbs",
    "user_agents.txt"
)

BLOCKED_WAIT_TIMEOUT = 2

FIRST_LEVEL_BORDER = .05
SECOND_LEVEL_BORDER = .15
THIRD_LEVEL_BORDER = .3

ITERATION_SIZE = 300

MAX_FILTER_SIZE = 1e7


def error_callback_pass():
    pass


def error_callback_raise():
    raise requests.HTTPError("Bad return code")


def queue_timeout_wait():
    time.sleep(BLOCKED_WAIT_TIMEOUT)


def queue_timeout_quit():
    from cobweb import Cobweb
    Cobweb().stop()


ERROR_CALLBACKS = {
    "skip": error_callback_pass,
    "raise": error_callback_raise,

}

QUEUE_TIMEOUT_CALLBACKS = {
    "wait": queue_timeout_wait,
    "quit": queue_timeout_quit
}



import logging
from io import BytesIO

from cobweb.adapters.session import SessionAbstractAdapter
from cobweb.bbs import BlockBypassSystem


class FileAdapter(SessionAbstractAdapter):

    bbs = BlockBypassSystem()

    def __init__(self, **kwargs):
        super(FileAdapter, self).__init__(**kwargs)

    @property
    def logger(self):
        return logging.getLogger("cobweb.adapters.file")

    @bbs.monitor_responses
    def process_response(self, response):
        response = super().process_response(response)
        response.processed_by = self.__class__

        response.data = BytesIO(response.content)
        return response



import json
import logging
import os
from types import FunctionType

import yaml

from cobweb.defaults import ERROR_CALLBACKS, QUEUE_TIMEOUT_CALLBACKS


class Settings:

    ERROR_SKIP = "skip"

    QUEUE_WAIT = "wait"

    whitelist = {
        "use_bbs",
        "keep_alive_timeout",
        "page_load_timeout",
        "_success_codes",
        "_failure_codes",
        "on_failure_action",
        "on_failure_callback",
        "on_queue_timeout_action",
        "on_queue_timeout_callback",
    }

    __slots__ = whitelist

    def __init__(
            self,
            use_bbs=True,
            keep_alive_timeout=5,
            page_load_timeout=3,
            success_codes=frozenset({200, 202}),
            failure_codes=frozenset({400, 402, 403, 500, 502}),
            on_failure_action=ERROR_SKIP,
            on_queue_timeout_action=QUEUE_WAIT,
            on_failure_callback: FunctionType = None,
            on_queue_timeout_callback: FunctionType = None,
    ):

        self.use_bbs = use_bbs
        self.keep_alive_timeout = keep_alive_timeout
        self.page_load_timeout = page_load_timeout
        self._success_codes = set(success_codes)
        self._failure_codes = set(failure_codes)
        self.on_failure_action = on_failure_action
        self.on_failure_callback = on_failure_callback
        self.on_queue_timeout_action = on_queue_timeout_action
        self.on_queue_timeout_callback = on_queue_timeout_callback

    @property
    def logger(self):
        return logging.getLogger("cobweb.settings")

    @property
    def success_codes(self):
        return self._success_codes

    @success_codes.setter
    def success_codes(self, value):
        self._success_codes = set(value)

    @property
    def failure_codes(self):
        return self._failure_codes

    @failure_codes.setter
    def failure_codes(self, value):
        self._failure_codes = set(value)

    @property
    def attributes(self):
        return {attr: getattr(self, attr) for attr in self.whitelist}

    def default_error_handler(self):
        if self.on_failure_callback:
            return self.on_failure_callback
        else:
            return ERROR_CALLBACKS[self.on_failure_action]

    def default_queue_timeout_handler(self):
        if self.on_queue_timeout_callback:
            return self.on_queue_timeout_callback
        else:
            return QUEUE_TIMEOUT_CALLBACKS[self.on_queue_timeout_action]

    def update(self, data: dict) -> None:
        for key, value in data.items():
            setattr(self, key, value)

    def to_json(self, filename: str) -> None:
        with open(Settings.get_absolute_path(filename), "w") as file:
            json.dump(self.attributes, file)

    def to_yaml(self, filename: str) -> None:
        with open(Settings.get_absolute_path(filename), "w") as file:
            yaml.dump(self.attributes, file, default_flow_style=False)

    @classmethod
    def from_json(cls, filename: str) -> 'Settings':
        settings = Settings()
        with open(Settings.get_absolute_path(filename)) as file:
            settings_data = Settings.sanitize(json.load(file), cls.whitelist)
            settings.update(settings_data)
        return settings

    @classmethod
    def from_yaml(cls, filename: str) -> 'Settings':
        settings = Settings()
        with open(Settings.get_absolute_path(filename)) as file:
            settings_data = Settings.sanitize(yaml.load(file), cls.whitelist)
            settings.update(settings_data)
        return settings

    @staticmethod
    def sanitize(data: dict, whitelist: set) -> dict:
        return {key: value for key, value in data.items() if key in whitelist}

    @staticmethod
    def get_absolute_path(relative_path: str) -> str:
        return os.path.abspath(
            os.path.expandvars(
                os.path.expanduser(
                    relative_path
                )
            )
        )



from abc import abstractmethod, ABCMeta


class AbstractAdapter(metaclass=ABCMeta):

    @abstractmethod
    def invoke(self, url: str):
        raise NotImplementedError("Method invocation on an abstract class")

    @abstractmethod
    def send(self, url: str):
        raise NotImplementedError("Method invocation on an abstract class")

    @abstractmethod
    def prepare_request(self, url: str):
        raise NotImplementedError("Method invocation on an abstract class")

    @abstractmethod
    def process_response(self, response):
        raise NotImplementedError("Method invocation on an abstract class")



import logging

from bs4 import BeautifulSoup
from requests import HTTPError

from cobweb.adapters.session import SessionAbstractAdapter
from cobweb.bbs import BlockBypassSystem


class HttpAdapter(SessionAbstractAdapter):

    bbs = BlockBypassSystem()

    def __init__(self, **kwargs):

        super(HttpAdapter, self).__init__(**kwargs)

        self.accept_content_types = {
            "text/html",
            "text/plain",
            "application/xhtml+xml",
            "application/xml",
        }

    @property
    def logger(self):
        return logging.getLogger("cobweb.adapters.http")

    @bbs.monitor_responses
    def process_response(self, response):
        response = super().process_response(response)

        if response.headers["content-type"] not in self.accept_content_types:
            raise HTTPError("Wrong content type")

        response.processed_by = self.__class__
        response.data = BeautifulSoup(response.content, "html.parser")
        return response



from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.ui import WebDriverWait

from cobweb.adapters.http import HttpAdapter


class DynamicHttpAdapter(HttpAdapter):

    PhantomJSDriver = webdriver.PhantomJS
    ChromeDriver = webdriver.Chrome
    FirefoxDriver = webdriver.Firefox
    SafariDriver = webdriver.Safari

    def __init__(self, webdriver=PhantomJSDriver, sample_id="root", **kwargs):
        super(DynamicHttpAdapter, self).__init__(**kwargs)

        self.sample_id = sample_id
        self.driver = webdriver

    def __enter__(self):
        self.driver()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.driver.close()

    def invoke(self, url):
        return self.send(url)

    def send(self, url):
        url = self.prepare_request(url)
        self.driver().get(url)
        wait = WebDriverWait(self.driver, 10)
        wait.until(expected_conditions.visibility_of_element_located((By.TAG_NAME, "body")))
        return self.process_response(self.driver)

    def prepare_request(self, url: str):
        return url

    def process_response(self, response):
        data = BeautifulSoup(response.page_source, "html.parser")
        cookies = response.get_cookies()
        return {"data": data, "cookies": cookies}



import functools
import multiprocessing
import threading
from collections import deque

import requests


def is_connected_to_internet():
    try:
        requests.get('http://216.58.192.142', timeout=2)
        return True
    except requests.exceptions.ConnectionError:
        return False


class Singleton(type):

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Rectifier(threading.Thread, metaclass=Singleton):

    def __init__(self):
        threading.Thread.__init__(self)
        self.queue = deque()
        self.process_lock = multiprocessing.Lock()

    def run(self):
        if not self.queue:
            raise ValueError("No queue to take tasks from")
        while True:
            try:
                task = self.queue.pop()
            except IndexError:
                pass
            else:
                self.process_lock.acquire()
                task[0](*task[1], **task[2])
                self.process_lock.release()


def sequential(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
        rectifier = Rectifier()
        rectifier.queue.appendleft((func, args, kwargs))
        if not rectifier.is_alive():
            rectifier.start()

    return inner



import argparse
import code
import os

from cobweb.adapters.http import HttpAdapter

EMPTY_PROJECT_CODE = """from cobweb import Cobweb

START_URL = "example.com"


def process_callback(response):
    pass


if __name__ == "__main__":
    cobweb = Cobweb()
    cobweb.start_from(START_URL)
    cobweb.default_process_function = process_callback
    cobweb.populate_spiders()
    cobweb.start()
"""


def main():
    parser = argparse.ArgumentParser(description='Cobweb helper line')
    parser.add_argument('command',
                        help='command to be executed',
                        choices=["init", "shell"],
                        type=str
                        )
    parser.add_argument("parameter",
                        help="parameter to the command",
                        type=str,
                        nargs="?",
                        )
    args = parser.parse_args()

    if args.command == "init":
        if not args.parameter:
            print("No project name")
        else:
            init(args.parameter)
    elif args.command == "shell":
        if not args.parameter:
            print("No url passed to shell")
        else:
            shell(args.parameter)


def init(name):
    current_directory = os.getcwd()
    project_name = os.path.join(current_directory, name)

    if os.path.exists(project_name):
        raise OSError("Folder with name '{}' already exists".format(name))

    os.makedirs(os.path.join(project_name, "adapters"))
    os.makedirs(os.path.join(project_name, "spiders"))
    with open(os.path.join(project_name, "__main__.py"), "w") as file:
        file.write(EMPTY_PROJECT_CODE)


def shell(url):
    response = HttpAdapter().invoke(url)
    code.interact(local=dict(globals(), **locals()))



import logging
import multiprocessing
from queue import Empty
from types import FunctionType
from urllib.parse import unquote

from requests import ReadTimeout

from cobweb.adapters import AbstractAdapter
from cobweb.adapters.http import HttpAdapter
from cobweb.link_extractor import LinkExtractor
from cobweb.settings import Settings


class Spider(multiprocessing.Process):

    NORMAL = 0

    def __init__(
            self,
            queue=None,
            settings: Settings = Settings(),
            adapter: AbstractAdapter = HttpAdapter(),
            link_extractor: LinkExtractor = LinkExtractor(),
            process_response: FunctionType = None,
            on_failure_callback: FunctionType = None,
            on_queue_timeout_callback: FunctionType = None,
            validator: FunctionType = None):

        multiprocessing.Process.__init__(self)

        self._queue = queue
        self._link_extractor = link_extractor
        self._settings = settings
        self._adapter = adapter
        self._on_failure_callback = None
        self._on_queue_timeout_callback = None
        self._process_response = None
        self._validator = None

        self.daemon = False
        self.keep_alive = True
        self.process_response = process_response
        self.on_failure_callback = on_failure_callback
        self.on_queue_timeout = on_queue_timeout_callback
        self.validator = validator

    @property
    def logger(self):
        return logging.getLogger('cobweb.spiders')

    @property
    def adapter(self):
        return self._adapter

    @property
    def on_failure_callback(self):
        if self._on_failure_callback:
            return self._on_failure_callback
        else:
            return self._settings.default_error_handler

    @on_failure_callback.setter
    def on_failure_callback(self, value: FunctionType):
        if not isinstance(value, FunctionType) and value is not None:
            raise ValueError("Trying to set not a function to callback")
        else:
            self._on_failure_callback = value

    @property
    def on_queue_timeout(self):
        if self._on_queue_timeout_callback:
            return self._on_queue_timeout_callback
        else:
            return self._settings.default_queue_timeout_handler

    @on_queue_timeout.setter
    def on_queue_timeout(self, value: FunctionType):
        if not isinstance(value, FunctionType) and value is not None:
            raise ValueError("Trying to set not a function to callback")
        else:
            self._on_queue_timeout_callback = value

    @property
    def process_response(self):
        return self._process_response

    @process_response.setter
    def process_response(self, value: FunctionType):
        if not isinstance(value, FunctionType) and value is not None:
            raise ValueError("Trying to set not a function to callback")
        else:
            self._process_response = value

    @property
    def validator(self):
        return self._validator

    @validator.setter
    def validator(self, value: FunctionType):
        if not isinstance(value, FunctionType) and value is not None:
            raise ValueError("Trying to set not a function to callback")
        else:
            self._validator = value

    @property
    def queue(self):
        return self._queue

    @queue.setter
    def queue(self, value):
        self._queue = value

    @property
    def keep_alive_timeout(self):
        return self._settings.keep_alive_timeout

    @property
    def block_level(self):
        return self._settings.block_level

    @property
    def success_codes(self):
        return self._settings.success_codes

    @property
    def failure_codes(self):
        return self._settings.failure_codes

    @property
    def settings(self):
        raise AttributeError("Settings object isn't directly accessible")

    @settings.setter
    def settings(self, value: Settings):
        self._settings = value

    def is_alive(self):
        return self.keep_alive

    def stop(self):
        self.keep_alive = False

    def _extract_links_async(self, response):
        if response.headers["content-type"] == "text/html":
            self._link_extractor.responses_queue.put(response)

    def _process_response_async(self, response):
        self.process_response(response)

    def _process_url(self, url: str) -> None:
        with self.adapter:
            response = self.adapter.invoke(url)

            if response.status_code in self.success_codes:
                if self.validator and not self.validator(response):
                    return

                self._extract_links_async(response)
                self._process_response_async(response)
            elif response.status_code in self.failure_codes:
                self.on_failure_callback(response)
            else:
                # What to do here?
                pass

    def run(self):
        if not self.process_response:
            raise ValueError("No processing function specified")

        self.logger.info("starting queue extraction loop")
        while self.keep_alive:
            try:
                url = self.queue.get(timeout=self.keep_alive_timeout)
            except Empty:
                self.logger.warning(
                    "queue is empty: waited for %s",
                    self.keep_alive_timeout
                )
                self.on_queue_timeout()
                self.logger.debug("queue timeout callback was called. continue execution")
            else:
                self.logger.debug("fetched new url from queue: %s", unquote(url))
                self.logger.debug("queue current size is %s", self.queue.qsize())
                try:
                    self._process_url(url)
                except ReadTimeout:
                    pass
                    # What to do when internet slow?
                else:
                    self.logger.info("page at %s was successfully processed", unquote(url))



from setuptools import setup

setup(
    name='cobweb-scrapper',
    version='1.0.0',
    packages=['cobweb', 'cobweb.spiders', 'cobweb.bbs', 'cobweb.helpers',
              'cobweb.adapters', 'cobweb.link_extractor'],
    url='',
    license='',
    author='Alexander Svito',
    author_email='alexandervirk@gmail.com',
    entry_points={
        'console_scripts': ['cobweb=cobweb.scripts:main'],
    },
    data_files=[('cobweb/bbs', ['cobweb/bbs/user_agents.txt'])],
    description='A package for easy scrapping', requires=['requests', 'PyYAML', 'pygtrie', 'bs4', 'urllib3', 'Pillow',
                                                          'selenium']
)



from abc import ABCMeta

from requests import Session, Request
from requests.cookies import RequestsCookieJar

from cobweb.adapters import AbstractAdapter
from cobweb.bbs import BlockBypassSystem


class SessionAbstractAdapter(AbstractAdapter, metaclass=ABCMeta):

    bbs = BlockBypassSystem()

    def __init__(
            self,
            use_bbs=True,
            method="GET",
            stream=False,
            proxies=None,
            ssl_certificate=None,
            keep_session=True,
            timeout=(10, 10),
            data=None,
            headers=None):

        if not use_bbs:
            self.bbs.stop()

        self._use_bbs = use_bbs
        self._headers = {"User-Agent": "Python-Cobweb/1.0.0"}
        if headers:
            self.headers.update(**headers)
        self._data = data
        self._timeout = timeout
        self._ssl_certificate = ssl_certificate
        self._proxies = proxies
        self._method = method
        self._stream = stream
        self._response = None
        self._content = None
        self._cookies = None
        self.session = Session() if keep_session else None
        self.accept_content_types = {}

    @property
    def use_bbs(self):
        return self._use_bbs

    @property
    def logger(self):
        raise ValueError("Called logger on an abstract class")

    @property
    def headers(self) -> dict:
        return self._headers

    @property
    def cookies(self) -> RequestsCookieJar:
        return self._cookies

    @use_bbs.setter
    def use_bbs(self, value):
        self._use_bbs = value
        if not value:
            self.bbs.stop()
        else:
            self.bbs.start()

    def __enter__(self):
        self.session = Session()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_session()

    def invoke(self, url: str):
        if not self.session:
            self.logger.debug("open session")
            self.session = Session()
            response = self.send(url)
            self.close_session()
        else:
            response = self.send(url)
        return response

    def close_session(self):
        self.session.close()
        self.session = None
        self.logger.debug("session closed")

    def send(self, url: str):
        self.logger.debug("enter session")
        with self.session as session:
            response = session.send(
                self.prepare_request(url),
                stream=self._stream,
                verify=self._ssl_certificate,
                proxies=self._proxies,
                cert=self._ssl_certificate,
                timeout=self._timeout
            )
        self.logger.debug("got response")
        return self.process_response(response)

    @bbs.monitor_requests
    def prepare_request(self, url: str):
        self.logger.debug("preparing request")
        request = Request(self._method, url, data=self._data, headers=self._headers)
        self._cookies = request.cookies
        return self.session.prepare_request(request)

    def process_response(self, response):
        if ";" in response.headers["content-type"]:
            response.headers["content-type"] = response.headers["content-type"].partition(";")[0]
        return response



import functools
import time
from random import choice

from cobweb.defaults import (
    BLOCKED_WAIT_TIMEOUT,
    USER_AGENT_PATH,
    FIRST_LEVEL_BORDER,
    SECOND_LEVEL_BORDER,
    ITERATION_SIZE,
    THIRD_LEVEL_BORDER
)


class BlockBypassSystem:

    ZERO = 0
    FIRST = 1
    SECOND = 2
    THIRD = 3

    def __init__(self, user_agents_filename=USER_AGENT_PATH):
        self._accepted_requests = 0
        self._blocked_request = 0
        self._total_requests = 0
        self._iteration = ITERATION_SIZE
        self._blocked_wait_timeout = BLOCKED_WAIT_TIMEOUT
        self._is_active = True

        self.request_method_order = [
            [],
            ["modify_user_agent"],
            ["wait_timeout"],
            []
        ]

        self.response_method_order = [
            ["count_statistics"],
            ["remove_honeypot_links"],
            [],
            []
        ]
        self._honeypot_content_types = {"text/html", }

        with open(user_agents_filename) as file:
            self.headers_list = file.readlines()

    @property
    def block_level(self):
        blocked_percent = self._accepted_requests / self._total_requests
        if blocked_percent < FIRST_LEVEL_BORDER:
            return self.ZERO
        elif blocked_percent < SECOND_LEVEL_BORDER:
            return self.FIRST
        elif blocked_percent < THIRD_LEVEL_BORDER:
            return self.SECOND
        else:
            return self.THIRD

    @property
    def accepted_requests(self):
        return self._accepted_requests

    @accepted_requests.setter
    def accepted_requests(self, value):
        self._total_requests += value - self._accepted_requests
        self._accepted_requests = value
        self.check_iteration()

    @property
    def blocked_request(self):
        return self._blocked_request

    @blocked_request.setter
    def blocked_request(self, value):
        self._total_requests += value - self._blocked_request
        self._blocked_request = value
        self.check_iteration()

    def stop(self):
        self._is_active = False

    def start(self):
        self._is_active = True

    def check_iteration(self):
        if self._total_requests % self._iteration == 0:
            self._total_requests = 0
            self._blocked_request = 0
            self._accepted_requests = 0

    def monitor_responses(self, method):
        @functools.wraps(method)
        def inner(*args, **kwargs):
            response = method(*args, **kwargs)
            if self._is_active:
                return self.check_response(response)
            else:
                return response
        return inner

    def monitor_requests(self, method):
        @functools.wraps(method)
        def inner(*args, **kwargs):
            request = method(*args, **kwargs)
            if self._is_active:
                return self.take_action(request)
            else:
                return request
        return inner

    def count_statistics(self, response):
        if response.ok:
            self.accepted_requests += 1
        else:
            self.blocked_request += 1
        return response

    def check_response(self, response):
        for actions in reversed(self.response_method_order):
            for action in actions:
                response = getattr(self, action)(response)
        return response

    def remove_honeypot_links(self, response):
        if response.headers["content-type"] in self._honeypot_content_types:
            for link in response.data.select("a[style~='visibility:hidden']"):
                link.decompose()
            for link in response.data.select("a[style~='display:none']"):
                link.decompose()
        return response

    def take_action(self, request):
        for actions in reversed(self.request_method_order):
            for action in actions:
                request = getattr(self, action)(request)
        return request

    def modify_user_agent(self, request):
        user_agent = choice(self.headers_list)
        request.headers["user-agent"] = user_agent.strip()
        return request

    def wait_timeout(self, request):
        time.sleep(self._blocked_wait_timeout)
        return request
